@@ -1,4 +1,4 @@
-from typing import List
+from typing import List, Tuple
 
 import numpy as onp
 from jax import random, nn, lax, jit, numpy as np
@@ -41,8 +41,7 @@ def apply_rotary_pos_emb(x, sincos):
     sin, cos = sincos
     return (x * cos) + (rotate_every_two(x) * sin)
 
-# feedforward
-# SwiGLU variant
+# swish for SwiGLU variant
 
 def swish(x):
     return x * nn.sigmoid(x)
@@ -67,17 +66,17 @@ class FeedForward(Module):
         x *= swish(gate)
         return x @ self.wo
 
-# attention
-# multi-query, one-headed key / values variant
+# attention - multi-query, one-headed key / values variant
+# feedforward - Shazeer's SwiGLU variant
 
-class Attention(Module):
+class ParallelTransformerBlock(Module):
     norm: Module
-    wq: np.ndarray
-    wk: np.ndarray
-    wv: np.ndarray
-    wo: np.ndarray
+    wi: np.ndarray
+    attn_wo: np.ndarray
+    ff_wo: np.ndarray
 
     heads: int = static_field()
+    fused_dims: Tuple[int] = static_field()
     scale: float = static_field()
     mask_value: float = static_field()
 
@@ -87,28 +86,30 @@ class Attention(Module):
         dim_head,
         heads,
         key,
+        ff_mult = 4,
         mask_value = -1e10
     ):
-        inner_dim = dim_head * heads
+        attn_inner_dim = dim_head * heads
+        ff_inner_dim = dim * ff_mult
         self.norm = LayerNorm(dim)
+        self.fused_dims = (attn_inner_dim, dim_head, dim_head, ff_inner_dim, ff_inner_dim)
 
-        self.wq = random.normal(key, (dim, inner_dim))
-        self.wk = random.normal(key, (dim, dim_head))
-        self.wv = random.normal(key, (dim, dim_head))
-        self.wo = random.normal(key, (inner_dim, dim))
+        self.wi = random.normal(key, (dim, sum(self.fused_dims)))
+        self.attn_wo = random.normal(key, (attn_inner_dim, dim))
+        self.ff_wo = random.normal(key, (ff_inner_dim, dim))
 
         self.heads = heads
         self.scale = dim_head ** -0.5
         self.mask_value = mask_value
 
     def __call__(self, x, *, pos_emb, causal_mask):
-        n = x.shape[-2]
+        n, split_indices = x.shape[-2], onp.cumsum(self.fused_dims[:-1])
 
         x = self.norm(x)
 
-        q = x @ self.wq
-        k = x @ self.wk
-        v = x @ self.wv
+        # fused attention and feedforward projections
+
+        q, k, v, ff, ff_gate = np.split(x @ self.wi, split_indices, axis = -1)
 
         # split out heads
 
@@ -142,9 +143,15 @@ class Attention(Module):
 
         out = rearrange(out, '... h n d -> ... n (h d)')
 
+        # feedforward out
+
+        attn_out = out @ self.attn_wo
+
+        ff_out = (ff * swish(ff_gate)) @ self.ff_wo
+
         # combine heads out
 
-        return out @ self.wo
+        return attn_out + ff_out
 
 # main class
 
@@ -168,25 +175,19 @@ class PaLM(Module):
         self.embedding = random.normal(key, (num_tokens, dim)) * 0.02
         self.inv_freq = 1.0 / (10000 ** (np.arange(0, dim_head, 2) / dim_head))
 
-        self.layers = []
-        for _ in range(depth):
-            attn = Attention(dim = dim, dim_head = dim_head, heads = heads, key = key)
-            ff = FeedForward(dim = dim, mult = ff_mult, key = key)
-            self.layers.append([attn, ff])
-
+        self.layers = [ParallelTransformerBlock(dim = dim, dim_head = dim_head, heads = heads, ff_mult = ff_mult, key = key) for _ in range(depth)]
         self.norm = LayerNorm(dim)
 
     @jit
     def __call__(self, x):
-        n = x.shape[-2]
+        n = x.shape[-1]
         x = self.embedding[x]
 
         rotary_emb = fixed_pos_embedding(self.inv_freq, n)
         causal_mask = np.tril(np.ones((n, n)))
 
-        for attn, ff in self.layers:
-            x = attn(x, pos_emb = rotary_emb, causal_mask = causal_mask) + x
-            x = ff(x) + x
+        for block in self.layers:
+            x = block(x, pos_emb = rotary_emb, causal_mask = causal_mask) + x
 
         x = self.norm(x)
         return x @ self.embedding.transpose()
